/*
 * @(#)SCApplication.cpp	1.00 10th June 2000
 * 
 * Contains functions to implement the sample application that SCTool
 * demonstrates.
 *
 * Modification Log:
 */

/**
 *
 * SCTool is a sample application using Schlumberger PayFlex 1K smart cards.
 * Using this application you can personalize a card, and store data either
 * protected with a pin number or without protection, as you desire.
 * 
 * Several generic functions have been written for SmartCards, so that they
 * can be ported later to any type of SmartCard. 
 * 
 * For more information on SmartCards and how to program them, refer to the
 * "Smart Card Crash Course" at my website: 
 * http://tanmaykm.tripod.com.
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 10th June, 2000
 */

#include <stdio.h>
#include <string.h>
#include "SCConnectTool.hxx"
#include "SCMessageTool.hxx"
#include "SCFileTool.hxx"
#include "SCAuthTool.hxx"
#include "SCApplication.hxx"

BYTE MASTER_FILE[2] = { 0x3f, 0x00 };

// Transport key is sensitive information.
// You should change the transport key after personalization.
// The transport key mentioned here may be incorrect.
// Transport keys are supplied by the card supplier to customers.
// You can obtain the transport key from your vendor.
BYTE TRANSPORT_KEY_0[8] = { 0x63, 0x90, 0x49, 0x38, 0x39, 0x55, 0x46, 0x47 };

BYTE USERID_FILE[2] = { 0x7f, 0x01 };
BYTE USERID_FILE_SIZE[2] = { 0x00, MAX_USER_INFO_SIZE }; // 64 bytes

BYTE SIZE_PIN_FILE[] = { 0x00, 0x0A };
BYTE SIZE_C_KEY[] = { 0x00, 0x0A };
BYTE SIZE_V_KEY[] = { 0x00, 0x0A };

BYTE ID_PIN_FILE[] = { 0x00, 0x00 };
BYTE ID_V_KEY_FILE[] = { 0x00, 0x01 };
BYTE ID_C_KEY_FILE[] = { 0x00, 0x10 };

BYTE FILETYPE_C_KEY = 0x12;
BYTE FILETYPE_V_KEY = 0x32;
BYTE FILETYPE_PIN = 0x22;

// record length. 
// 1 byte activation indicator, 8 byte pin, 1 byte max attempt counter.
BYTE RL_PIN = 0x0A; 
BYTE RL_V_KEY = 0x0A;
BYTE RL_C_KEY = 0x0A;

// it's upto the program to do internal authentication
BYTE AC_DF[] = { 0x02, 0x00 }; // ext authentication (AUT) required to create file
BYTE AC_PIN[] = { 0xF0, 0x00 }; // read never, write always, unblock always
BYTE AC_V_KEY[] = { 0xF0, 0x00 }; // read never, write always, unblock always
BYTE AC_C_KEY[] = { 0xF0, 0x00 }; // read never, write always
BYTE AC_USER_DATA[] = { 0x33, 0x0F }; // ext auth and pin required for read and write, write can never be locked
BYTE AC_NONE[] = { 0x00, 0x00 };

BYTE KN_DF[] = { 0x00, 0x00 };    // 0th record of key
BYTE KN_PIN[] = { 0x00, 0x00 };   // 0th record of key
BYTE KN_V_KEY[] = { 0x00, 0x00 }; // 0th record of key
BYTE KN_C_KEY[] = { 0x00, 0x00 }; // 0th record of key
BYTE KN_USER_DATA[] = {0x00, 0x00 }; // 0th record of key
BYTE KN_NONE[] = { 0x00, 0x00 };

BYTE PLAIN_APP_DF[2] = { 0x7f, 0x00 };
BYTE PLAIN_APP_DF_SIZE[2] = { 0x00, 0x80 }; // 128 bytes

BYTE PIN_APP_DF[2] = { 0x7f, 0x30 }; // already used 7f10, 7f20
//BYTE PIN_APP_DF_SIZE[2] = { 0x00, 0xc8 }; // 200 bytes
BYTE PIN_APP_DF_SIZE[2] = { 0x00, 0x80 }; // 128 bytes

// Note:
// The terminology "Calculation Key" and "Verification Key" are relative.
// That is, for a program, a key is called a calculation key if it is used 
// to respond to a challenge. The same key is called verification key by
// the challenger.
// Typically these keys should NOT be same for all cards and
// should be generated from a card serial number.
BYTE ExtKey[] = {1,2,3,4,5,6,7,8}; // value of V Key stored in card used for external auth.
BYTE IntKey[] = {8,7,6,5,4,3,2,1}; // value of C Key stored in card used for internal auth.

unsigned char DEF_USER_INFO[] = "userid passw lang";

#ifdef _DEBUG
#define WAIT_FOR_ENTER { printf("Press Enter Key..."); fflush(stdin); fflush(stdout); getchar(); }
#else
#define WAIT_FOR_ENTER 
#endif

int UnblockChangePinInUserCard(int iUnblock, BYTE *OldPin, BYTE *NewPin, char *sMessage)
{
	SCARDCONTEXT hSC;
	SCARDHANDLE hCard;
	char sReader[1024], sCard[1024];
	long lSysErr;

	if (0 != ConnectCard(&hSC, &hCard, sReader, sCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		printf("%s\n", sMessage);
		return -1;
	}
	printf("Connected to card\n");
	WAIT_FOR_ENTER;

	PrintReaderCharacteristics(hSC, hCard);
	PrintCardStatus(hCard);

	if(0 != SelectFileAndCheckErr(hCard, MASTER_FILE, sMessage))
	{
		printf("%s\n", sMessage);
		goto DISCONNECT_AND_EXIT;
	}
	printf("Selected master file\n");
	WAIT_FOR_ENTER;

	if(0 != SelectFileAndCheckErr(hCard, PIN_APP_DF, sMessage))
	{
		printf("%s\n", sMessage);
		goto DISCONNECT_AND_EXIT;
	}
	printf("Selected App DF\n");
	WAIT_FOR_ENTER;

	// do internal authentication to verify the card
	// we use key number 0 because we wrote only one key
	if(0 != InternalAuthentication(hCard, 0x00, IntKey, sMessage))
	{
		printf("%s\n", sMessage);
		goto DISCONNECT_AND_EXIT;
	}
	printf("Internal authentication done\n");
	WAIT_FOR_ENTER;

	// do external authentication to verify the card
	// we use key number 0 because we wrote only one key
	if(0 != ExternalAuthentication(hCard, 0x00, ExtKey, sMessage))
	{
		printf("%s\n", sMessage);
		goto DISCONNECT_AND_EXIT;
	}
	printf("External authentication done\n");
	WAIT_FOR_ENTER;

	if(iUnblock)
	{
		// unblock PIN
		if(0 != UnblockKeyAndCheckErr(hCard, OldPin, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Unblocked Pin\n");
		WAIT_FOR_ENTER;
	}
	else
	{
		if(0 != ChangeKeyAndCheckErr(hCard, 1, OldPin, NewPin, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Changed Pin\n");
		WAIT_FOR_ENTER;
	}

DISCONNECT_AND_EXIT:
	if (0 != DisconnectCard(&hSC, &hCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}

	return 0;
}

int UnblockPinUserCard(BYTE *Pin, char *sMessage)
{
	return UnblockChangePinInUserCard(1, Pin, 0, sMessage);
}

int ChangePinInUserCard(BYTE *OldPin, BYTE *NewPin, char *sMessage)
{
	return UnblockChangePinInUserCard(0, OldPin, NewPin, sMessage);
}

int AccessPinUserCard(int iMode, BYTE *sData, BYTE DataLen, BYTE *Pin, char *sMessage)
{
	SCARDCONTEXT hSC;
	SCARDHANDLE hCard;
	char sReader[1024], sCard[1024];
	long lSysErr;
	int iRet = -1;

	if (0 != ConnectCard(&hSC, &hCard, sReader, sCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		printf("%s\n", sMessage);
		return -1;
	}
	printf("Connected to card\n");
	WAIT_FOR_ENTER;

	PrintReaderCharacteristics(hSC, hCard);
	PrintCardStatus(hCard);

	if(0 != SelectFileAndCheckErr(hCard, MASTER_FILE, sMessage))
	{
		printf("%s\n", sMessage);
		goto DISCONNECT_AND_EXIT;
	}
	printf("Selected master file\n");
	WAIT_FOR_ENTER;

	PrintCardUsageInfo(hCard, sMessage);

	if(CREATE_MODE == iMode)
	{
		// verify with transport key
		if(0 != VerifyKeyAndCheckErr(hCard, 0, TRANSPORT_KEY_0, 8, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Verified transport key\n");
		WAIT_FOR_ENTER;

		// create the app DF
		if(0 != CreateDFAndCheckErr(hCard, PIN_APP_DF, PIN_APP_DF_SIZE, AC_DF, KN_DF, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Created app DF\n");
		WAIT_FOR_ENTER;

		if(0 != SelectFileAndCheckErr(hCard, PIN_APP_DF, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected app DF\n");
		WAIT_FOR_ENTER;

		// verify with transport key
		if(0 != VerifyKeyAndCheckErr(hCard, 0, TRANSPORT_KEY_0, 8, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Verified transport key\n");
		WAIT_FOR_ENTER;

		// create pin file
		if(0 != CreateEFAndCheckErr(hCard, ID_PIN_FILE, SIZE_PIN_FILE, AC_PIN, KN_PIN, RL_PIN, FILETYPE_PIN, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Created pin file\n");
		WAIT_FOR_ENTER;

		// create c file
		if(0 != CreateEFAndCheckErr(hCard, ID_C_KEY_FILE, SIZE_C_KEY, AC_C_KEY, KN_C_KEY, RL_C_KEY, FILETYPE_C_KEY, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Created calculation key file\n");
		WAIT_FOR_ENTER;

		// create v file
		if(0 != CreateEFAndCheckErr(hCard, ID_V_KEY_FILE, SIZE_V_KEY, AC_V_KEY, KN_V_KEY, RL_V_KEY, FILETYPE_V_KEY, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Created verification key file\n");
		WAIT_FOR_ENTER;

		// create user file
		if(0 != CreateETFAndCheckErr(hCard, USERID_FILE, USERID_FILE_SIZE, AC_USER_DATA, KN_USER_DATA, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Created user data file\n");
		WAIT_FOR_ENTER;

		// select pin file
		if(0 != SelectFileAndCheckErr(hCard, ID_PIN_FILE, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected app df\n");
		WAIT_FOR_ENTER;

		// write pin
		{
			BYTE PinData[10];
			PinData[0] = 0x0f; // 0=>key activated, 1=>key blocked, 15 remaining attempts
			for(int iIndex=0; iIndex<8; iIndex++) PinData[iIndex+1] = Pin[iIndex];
			PinData[9] = 0x0f; // 15 max failed verifications before key is blocked
			if(0 != WriteEFRecAndCheckErr(hCard, PinData, 10, sMessage))
			{
				printf("%s\n", sMessage);
				goto DISCONNECT_AND_EXIT;
			}
			printf("Wrote pin\n");
			WAIT_FOR_ENTER;
		}

		// select c key
		if(0 != SelectFileAndCheckErr(hCard, ID_C_KEY_FILE, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected calculation key file\n");
		WAIT_FOR_ENTER;

		// write c key
		{
			BYTE PinData[10];
			PinData[0] = 0x00; // key activated, single DES algo, version 0
			for(int iIndex=0; iIndex<8; iIndex++) PinData[iIndex+1] = IntKey[iIndex];
			PinData[9] = 0x01; // will be used for internal auth only
			if(0 != WriteEFRecAndCheckErr(hCard, PinData, 10, sMessage))
			{
				printf("%s\n", sMessage);
				goto DISCONNECT_AND_EXIT;
			}
			printf("Wrote Internal Key into C Key File\n");
			WAIT_FOR_ENTER;
		}

		// select v key
		if(0 != SelectFileAndCheckErr(hCard, ID_V_KEY_FILE, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected verification key file\n");
		WAIT_FOR_ENTER;

		// write v key
		{
			BYTE PinData[10];
			PinData[0] = 0x0f; // key activated, single DES algo, version 0, remaining attempts 15
			for(int iIndex=0; iIndex<8; iIndex++) PinData[iIndex+1] = ExtKey[iIndex];
			PinData[9] = 0x0f; // 15 max failed verifications before key is blocked
			if(0 != WriteEFRecAndCheckErr(hCard, PinData, 10, sMessage))
			{
				printf("%s\n", sMessage);
				goto DISCONNECT_AND_EXIT;
			}
			printf("Wrote External Key into V Key File\n");
			WAIT_FOR_ENTER;
		}

		if(0 != SelectFileAndCheckErr(hCard, USERID_FILE, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected user data file\n");
		WAIT_FOR_ENTER;

		// do internal authentication to verify the card
		// we use key number 0 because we wrote only one key
		if(0 != InternalAuthentication(hCard, 0x00, IntKey, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Internal authentication done\n");
		WAIT_FOR_ENTER;

		// do external authentication to let the card verify us
		// we use key number 0 because we wrote only one key
		if(0 != ExternalAuthentication(hCard, 0x00, ExtKey, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("External authentication done\n");
		WAIT_FOR_ENTER;

		// verify PIN
		if(0 != VerifyKeyAndCheckErr(hCard, 1, Pin, 8, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Verified Pin\n");
		WAIT_FOR_ENTER;

		// write user data
		if(0 != WriteETFAndCheckErr(hCard, sData, DataLen, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Wrote user data file\n");
		WAIT_FOR_ENTER;
	}
	else
	{
		if(0 != SelectFileAndCheckErr(hCard, PIN_APP_DF, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected App DF\n");
		WAIT_FOR_ENTER;

		// select user id file
		if(0 != SelectFileAndCheckErr(hCard, USERID_FILE, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Selected user id file\n");
		WAIT_FOR_ENTER;

		// do internal authentication to verify the card
		// we use key number 0 because we wrote only one key
		if(0 != InternalAuthentication(hCard, 0x00, IntKey, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Internal authentication done\n");
		WAIT_FOR_ENTER;

		// do external authentication to let the card verify us
		// we use key number 0 because we wrote only one key
		if(0 != ExternalAuthentication(hCard, 0x00, ExtKey, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("External authentication done\n");
		WAIT_FOR_ENTER;

		// verify PIN
		if(0 != VerifyKeyAndCheckErr(hCard, 1, Pin, 8, sMessage))
		{
			printf("%s\n", sMessage);
			goto DISCONNECT_AND_EXIT;
		}
		printf("Verified Pin\n");
		WAIT_FOR_ENTER;

		if(WRITE_MODE == iMode)
		{
			// write user id file
			if(0 != WriteETFAndCheckErr(hCard, sData, DataLen, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}
		else
		{
			// read user id file
			if(0 != ReadETFAndCheckErr(hCard, sData, DataLen, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}
	}

	iRet = 0;

DISCONNECT_AND_EXIT:
	if (0 != DisconnectCard(&hSC, &hCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}

	return iRet;
}

int AccessPlainUserCard(int iMode, BYTE *sData, BYTE DataLen, char *sMessage)
{
	SCARDCONTEXT hSC;
	SCARDHANDLE hCard;
	char sReader[1024], sCard[1024];
	long lSysErr;
	int iRet = -1;

	if (0 != ConnectCard(&hSC, &hCard, sReader, sCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}

	//PrintReaderCharacteristics(hSC, hCard);
	//PrintCardStatus(hCard);

	if(0 != SelectFileAndCheckErr(hCard, MASTER_FILE, sMessage))
	{
		goto DISCONNECT_AND_EXIT;
	}

	if(CREATE_MODE == iMode)
	{
		if(0 != VerifyKeyAndCheckErr(hCard, 0, TRANSPORT_KEY_0, 8, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}

		if(0 != SelectFileAndCheckErr(hCard, PLAIN_APP_DF, sMessage))
		{
			if(0 != CreateDFAndCheckErr(hCard, PLAIN_APP_DF, PLAIN_APP_DF_SIZE, AC_NONE, KN_NONE, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
			if(0 != SelectFileAndCheckErr(hCard, PLAIN_APP_DF, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}

		if(0 != VerifyKeyAndCheckErr(hCard, 0, TRANSPORT_KEY_0, 8, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}

		if(0 != SelectFileAndCheckErr(hCard, USERID_FILE, sMessage))
		{
			if(0 != CreateETFAndCheckErr(hCard, USERID_FILE, USERID_FILE_SIZE, AC_NONE, KN_NONE, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}

			if(0 != SelectFileAndCheckErr(hCard, USERID_FILE, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}

		// write user id file
		if(0 != WriteETFAndCheckErr(hCard, sData, DataLen, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}
	}
	else
	{
		if(0 != SelectFileAndCheckErr(hCard, PLAIN_APP_DF, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}

		if(0 != SelectFileAndCheckErr(hCard, USERID_FILE, sMessage))
		{
			goto DISCONNECT_AND_EXIT;
		}

		if(WRITE_MODE == iMode)
		{
			// write user id file
			if(0 != WriteETFAndCheckErr(hCard, sData, DataLen, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}
		else
		{
			// read user id file
			if(0 != ReadETFAndCheckErr(hCard, sData, DataLen, sMessage))
			{
				goto DISCONNECT_AND_EXIT;
			}
		}
	}

	iRet = 0;

DISCONNECT_AND_EXIT:
	if (0 != DisconnectCard(&hSC, &hCard, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}

	return iRet;
}
