/*
 * @(#)SCAuthTool.cpp	1.00 10th June 2000
 * 
 * Functions that take care of user authorization and security features
 * of smart cards.
 *
 * Modification Log:
 */

/**
 *
 * SCTool is a sample application using Schlumberger PayFlex 1K smart cards.
 * Using this application you can personalize a card, and store data either
 * protected with a pin number or without protection, as you desire.
 * 
 * Several generic functions have been written for SmartCards, so that they
 * can be ported later to any type of SmartCard. 
 * 
 * For more information on SmartCards and how to program them, refer to the
 * "Smart Card Crash Course" at my website: 
 * http://tanmaykm.tripod.com.
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 10th June, 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "SCConnectTool.hxx"
#include "SCAuthTool.hxx"
#include "SCMessageTool.hxx"

typedef struct 
{
    BYTE
        bCla,   // The instruction class
        bIns,   // The instruction code 
        bP1,    // Parameter to the instruction
        bP2,    // Parameter to the instruction
        bP3;    // Size of I/O Transfer
	BYTE abData[1024];
} APDUStruct;

extern "C"
{
	void key_setup(BYTE *key, BYTE *subkey);
	void block_cipher(BYTE *subkey, BYTE *text, int deciphering);
};

int VerifyKeyAndCheckErr(SCARDHANDLE hCard, int iIsCHVPin, BYTE *pKey, BYTE KeyLen, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != VerifyKey(hCard, iIsCHVPin, pKey, KeyLen, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int VerifyKey(SCARDHANDLE hCardHandle, int iIsCHVPin, BYTE *pKey, BYTE KeyLen, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x20;
	stSend.bP1 = 0x00;
	stSend.bP2 = iIsCHVPin ? 0x10 : 0x00; // check whether it's a key file (FT=0x32) or CHV (PIN/PUK) file (FT=0x22)
	stSend.bP3 = KeyLen;

	for(int iIndex=0; iIndex<8; iIndex++)
	{
		stSend.abData[iIndex] = pKey[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 13, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int UnblockKeyAndCheckErr(SCARDHANDLE hCard, BYTE *pKey, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != UnblockKey(hCard, pKey, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int UnblockKey(SCARDHANDLE hCardHandle, BYTE *pKey, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x2c;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00; 
	stSend.bP3 = 0x08;

	for(int iIndex=0; iIndex<8; iIndex++)
	{
		stSend.abData[iIndex] = pKey[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 13, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int ChangeKeyAndCheckErr(SCARDHANDLE hCard, int iIsCHVPin, BYTE *pOldKey, BYTE *pNewKey, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != ChangeKey(hCard, iIsCHVPin, pOldKey, pNewKey, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int ChangeKey(SCARDHANDLE hCardHandle, int iIsCHVPin, BYTE *pOldKey, BYTE *pNewKey, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x24;
	stSend.bP1 = 0x00;
	stSend.bP2 = iIsCHVPin ? 0x10 : 0x00; // check whether it's a key file (FT=0x32) or CHV (PIN/PUK) file (FT=0x22)
	stSend.bP3 = 0x10;

	// send old key
	for(int iIndex=0; iIndex<8; iIndex++)
	{
		stSend.abData[iIndex] = pOldKey[iIndex];
	}
	// send new key
	for(iIndex=0; iIndex<8; iIndex++)
	{
		stSend.abData[iIndex+8] = pNewKey[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 21, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int GetChallenge(SCARDHANDLE hCardHandle, BYTE *Challenge, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x84;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00; 
	stSend.bP3 = 0x08;

	// Transmit the request.
	lRecvBytes = 10;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5, 
							NULL, (LPBYTE)stSend.abData, &lRecvBytes);

	if(SCARD_S_SUCCESS == (*pRet))
	{
		RecvStatus[0] = stSend.abData[lRecvBytes-2];
		RecvStatus[1] = stSend.abData[lRecvBytes-1];
		for(int iIndex=0; iIndex < 8; iIndex++)
		{
			Challenge[iIndex] = stSend.abData[iIndex];
		}
	}

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int SendExtAuthToCard(SCARDHANDLE hCardHandle, BYTE KeyNumber, BYTE *Challenge, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x82;
	stSend.bP1 = 0x00;
	stSend.bP2 = KeyNumber; 
	stSend.bP3 = 0x06;

	for(int iIndex=0; iIndex < 6; iIndex++)
	{
		stSend.abData[iIndex] = Challenge[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 11, 
							NULL, RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int SendIntAuthToCard(SCARDHANDLE hCardHandle, BYTE KeyNumber, BYTE *Challenge, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0x88;
	stSend.bP1 = 0x00;
	stSend.bP2 = KeyNumber; 
	stSend.bP3 = 0x08;

	for(int iIndex=0; iIndex < 8; iIndex++)
	{
		stSend.abData[iIndex] = Challenge[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 13, 
							NULL, RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int ExternalAuthentication(SCARDHANDLE hCardHandle, BYTE KeyNumber, BYTE *ExtKey, char *sMessage)
{
	BYTE Challenge[8];
	BYTE DESKey[128];
	BYTE RecvStatus[2];
	long lSysErr;

	key_setup(ExtKey, DESKey);

	if(0 != GetChallenge(hCardHandle, Challenge, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}

	block_cipher(DESKey, Challenge, 0);

	if(0 != SendExtAuthToCard(hCardHandle, KeyNumber, Challenge, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}

	return 0;
}


int InternalAuthentication(SCARDHANDLE hCardHandle, BYTE KeyNumber, BYTE *IntKey, char *sMessage)
{
	BYTE Challenge[] = {1,3,5,7,2,4,6,8};
	BYTE DESKey[128];
	BYTE RecvStatus[2];
	BYTE Reply[6];
	long lSysErr;

	key_setup(IntKey, DESKey);

	if(0 != SendIntAuthToCard(hCardHandle, KeyNumber, Challenge, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}

	block_cipher(DESKey, Challenge, 0);

	if(0 != GetResponse(hCardHandle, Reply, 6, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}

	for(int iIndex=0; (iIndex<6) && (Challenge[iIndex] == Reply[iIndex]); iIndex++);

	if(6 != iIndex)
	{
		sprintf(sMessage, "Internal Authentication failed. Response didn't match Challenge.\n");
		return -1;
	}
	return 0;
}


