/*
 * @(#)SCConnectTool.cpp	1.00 10th June 2000
 * 
 * Functions that take care of connecting to a card reader and accessing
 * the card. Since these functions are written using PC/SC APIs, they will
 * work for any PC/SC compliant smartcard reader.
 *
 * Modification Log:
 */

/**
 *
 * SCTool is a sample application using Schlumberger PayFlex 1K smart cards.
 * Using this application you can personalize a card, and store data either
 * protected with a pin number or without protection, as you desire.
 * 
 * Several generic functions have been written for SmartCards, so that they
 * can be ported later to any type of SmartCard. 
 * 
 * For more information on SmartCards and how to program them, refer to the
 * "Smart Card Crash Course" at my website: 
 * http://tanmaykm.tripod.com.
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 10th June, 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "SCConnectTool.hxx"
#include "SCMessageTool.hxx"

typedef struct 
{
    BYTE
        bCla,   // The instruction class
        bIns,   // The instruction code 
        bP1,    // Parameter to the instruction
        bP2,    // Parameter to the instruction
        bP3;    // Size of I/O Transfer
	BYTE abData[1024];
} APDUStruct;


// Return:
// 0 : Success
// -1 : Could not establish card context
// -2 : Could not get a card reader
int ConnectCard(SCARDCONTEXT *phSC, SCARDHANDLE *phCardHandle, 
				char *sReader, char *sCard, 
				long *plSysErr)
{
	long lReturn;

	lReturn = SCardEstablishContext(SCARD_SCOPE_USER, NULL, NULL, phSC);
	if (SCARD_S_SUCCESS != lReturn)
	{
		*plSysErr = lReturn;
		return -1;
	}

	if(0 != ChooseReader(*phSC, phCardHandle, sReader, sCard, plSysErr))
	{
		return -2;
	}

	return 0;
}

// Returns:
// 0: Success
// -1 : Error disconnecting card
// -2 : Error releasing card context
int DisconnectCard(SCARDCONTEXT *phSC, SCARDHANDLE *phCardHandle, long *plSysErr)
{
	if(*phCardHandle)
	{
		*plSysErr = SCardDisconnect(*phCardHandle, SCARD_LEAVE_CARD);
		if(SCARD_S_SUCCESS != *plSysErr)
		{
			return -1;
		}
		*phCardHandle = 0;
	}

	if(*phSC)
	{
		*plSysErr = SCardReleaseContext(*phSC);
		if(SCARD_S_SUCCESS != *plSysErr)
		{
			return -2;
		}
		*phSC = 0;
	}
	return 0;
}

// Returns
//  0 : Success
// -1 : No readers available
// -2 : Error listing readers
int GetReadersList(SCARDCONTEXT hSC, char sList[20][160], int *piCount)
{
	LPTSTR          pmszReaders = NULL;
	LPTSTR          pReader;
	LONG            lReturn;
	DWORD           cch = SCARD_AUTOALLOCATE;
	int				iRet=0;

	// Retrieve the list the readers.
	lReturn = SCardListReaders(hSC, NULL, (LPTSTR)&pmszReaders, &cch);

	switch(lReturn)
	{
		case SCARD_S_SUCCESS:
			pReader = pmszReaders;
			for(*piCount=0; (pReader && *pReader) && ((*piCount)<20); (*piCount)++)
			{
				strncpy(sList[*piCount], pReader, sizeof(sList[0])); // Store the value.
				sList[*piCount][sizeof(sList[0])-1] = 0;
				pReader = pReader + strlen(pReader) + 1; // Advance to the next value.
			}
			SCardFreeMemory(hSC, pmszReaders); // Free the memory.
			break;
#ifdef SCARD_E_NO_READERS_AVAILABLE
		case SCARD_E_NO_READERS_AVAILABLE:
			iRet = -1;
			break;
#endif
		default:
			iRet = -2;
			break;
	}

	return iRet;
}


// Use SCardUIDlgSelectCard on Win2000
// Retruns:
// 0 : Success
// -1 ; Failure
int ChooseReader(SCARDCONTEXT hSC, SCARDHANDLE *pCardHandle, char *sReader, char *sCard,  long *plSysErr)
{
	OPENCARDNAME    dlgStruct;
	char            szReader[1024], szCard[1024];
	LONG            lReturn;

	// Initialize the structure.
	memset(&dlgStruct, 0, sizeof (dlgStruct));
	memset(szReader, 0, sizeof(szReader));
	memset(szCard, 0, sizeof(szCard));
	dlgStruct.dwStructSize = sizeof (dlgStruct);
	dlgStruct.hSCardContext = hSC;
	dlgStruct.dwFlags = SC_DLG_FORCE_UI;
	dlgStruct.lpstrRdr = szReader;
	dlgStruct.nMaxRdr = 1024;
	dlgStruct.lpstrCard = szCard;
	dlgStruct.nMaxCard = 1024;
	dlgStruct.lpstrTitle = "Select Card";
	dlgStruct.nMaxCardNames = 20;
	dlgStruct.dwShareMode = SCARD_SHARE_SHARED;
	dlgStruct.dwPreferredProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1;

	if(sReader) *sReader=0;
	if(sCard) *sCard=0;
	// Display the select card dialog.
	lReturn = GetOpenCardName(&dlgStruct);
	if (SCARD_S_SUCCESS != lReturn)
	{
		*plSysErr = lReturn;
		return -1;
	}
	*pCardHandle = dlgStruct.hCardHandle;
	szReader[dlgStruct.nMaxRdr-1] = 0;
	szCard[dlgStruct.nMaxCard-1] = 0;
	if(sReader) sprintf(sReader, "%s", szReader);
	if(sCard) sprintf(sCard, "%s", szCard);

	return 0;
}

int GetResponse(SCARDHANDLE hCardHandle, BYTE *Data, BYTE DataLen, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xc0;
	stSend.bP1 = 0x00;
	stSend.bP2 = 00; 
	stSend.bP3 = DataLen;

	// Transmit the request.
	lRecvBytes = 2+DataLen;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5, 
							NULL, stSend.abData, &lRecvBytes);

	if(SCARD_S_SUCCESS == (*pRet))
	{
		RecvStatus[0] = stSend.abData[lRecvBytes-2];
		RecvStatus[1] = stSend.abData[lRecvBytes-1];
		for(int iIndex=0; iIndex < DataLen; iIndex++)
		{
			Data[iIndex] = stSend.abData[iIndex];
		}
	}

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int PrintCardUsageInfo(SCARDHANDLE hCardHandle, char *sMessage)
{
	BYTE RecvStatus[2];
	BYTE Reply[28];
	long lSysErr;

	if(0 != GetResponse(hCardHandle, Reply, 28, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	printf("Available Memory: %xh %xh\n", Reply[0], Reply[1]);
	return 0;
}

int PrintReaderCharacteristics(SCARDCONTEXT hSC, SCARDHANDLE hCardHandle)
{
	SCTOOL_READER_CHARACTERISTICS stReader;

	GetReaderCharacteristics(hCardHandle, &stReader);

	if( (stReader.lVendorNameLen > 0) && (SCARD_AUTOALLOCATE != stReader.lATRLen) )
	{
		for(int iIndex=0; iIndex < (int)stReader.lVendorNameLen; iIndex++)
			printf("%c", *(stReader.pVendorName+iIndex));
		printf("\n");
	}

	if( (stReader.lATRLen > 0) && (SCARD_AUTOALLOCATE != stReader.lATRLen) )
	{
		printf("ATR: Length %ul bytes\n", stReader.lATRLen);
	}

	FreeReaderCharacteristics(hSC, &stReader);
	return 0;
}

void AllocAndGetAttrib(SCARDHANDLE hCardHandle, DWORD dwAttrib, unsigned long *pLen, LPBYTE pVal)
{
	*pLen = SCARD_AUTOALLOCATE;
	SCardGetAttrib(hCardHandle, dwAttrib, pVal, pLen);
}

void GetReaderCharacteristics(SCARDHANDLE hCardHandle, SCTOOL_READER_CHARACTERISTICS *pstReader)
{
	memset(pstReader, 0, sizeof(*pstReader));

	AllocAndGetAttrib(hCardHandle, SCARD_ATTR_ATR_STRING, &(pstReader->lATRLen), (LPBYTE)&pstReader->pATR);
	AllocAndGetAttrib(hCardHandle, SCARD_ATTR_VENDOR_NAME, &(pstReader->lVendorNameLen), (LPBYTE)&pstReader->pVendorName);
}

void FreeReaderCharacteristics(SCARDCONTEXT hSC, SCTOOL_READER_CHARACTERISTICS *pstReader)
{
	if(pstReader->pVendorName) SCardFreeMemory(hSC, pstReader->pVendorName);
	if(pstReader->pATR) SCardFreeMemory(hSC, pstReader->pATR);
}
