/*
 * @(#)SCFileTool.cpp	1.00 10th June 2000
 * 
 * Functions that take care of creating, reading and writing files into
 * a smartcard filesystem.
 *
 * Modification Log:
 */

/**
 *
 * SCTool is a sample application using Schlumberger PayFlex 1K smart cards.
 * Using this application you can personalize a card, and store data either
 * protected with a pin number or without protection, as you desire.
 * 
 * Several generic functions have been written for SmartCards, so that they
 * can be ported later to any type of SmartCard. 
 * 
 * For more information on SmartCards and how to program them, refer to the
 * "Smart Card Crash Course" at my website: 
 * http://tanmaykm.tripod.com.
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 10th June, 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "SCFileTool.hxx"
#include "SCMessageTool.hxx"

typedef struct 
{
    BYTE
        bCla,   // The instruction class
        bIns,   // The instruction code 
        bP1,    // Parameter to the instruction
        bP2,    // Parameter to the instruction
        bP3;    // Size of I/O Transfer
	BYTE abData[1024];
} APDUStruct;


int SelectFileAndCheckErr(SCARDHANDLE hCard, BYTE *fileId, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != SelectFile(hCard, fileId, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int SelectFile(SCARDHANDLE hCardHandle, BYTE *fileId, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xa4;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = 0x02;
	stSend.abData[0] = fileId[0];
	stSend.abData[1] = fileId[1];

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 7, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int CreateDFAndCheckErr(SCARDHANDLE hCard, BYTE *fileId, BYTE *FileSize, 
						BYTE *AccessCond, BYTE *KeyNum, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != CreateDFile(hCard, fileId, FileSize, AccessCond, KeyNum, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int CreateDFile(SCARDHANDLE hCardHandle, BYTE *fileId, BYTE *FileSize, 
				BYTE *AccessCond, BYTE *KeyNum,
				BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xe0;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = 0x09;
	stSend.abData[0] = FileSize[0];
	stSend.abData[1] = FileSize[1];
	stSend.abData[2] = fileId[0];
	stSend.abData[3] = fileId[1];
	stSend.abData[4] = 0x38; // file type
	stSend.abData[5] = AccessCond[0]; // access condition coding
	stSend.abData[6] = AccessCond[1];
	stSend.abData[7] = KeyNum[0]; // key number coding
	stSend.abData[8] = KeyNum[1];

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 14, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int CreateETFAndCheckErr(SCARDHANDLE hCard, BYTE *fileId, BYTE *FileSize, 
						BYTE *AccessCond, BYTE *KeyNum, 
						char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != CreateETFile(hCard, fileId, FileSize, AccessCond, KeyNum, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int CreateETFile(SCARDHANDLE hCardHandle, BYTE *fileId, BYTE *FileSize, 
				 BYTE *AccessCond, BYTE *KeyNum, 
				 BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xe0;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = 0x0B;
	stSend.abData[0] = FileSize[0];
	stSend.abData[1] = FileSize[1];
	stSend.abData[2] = fileId[0];
	stSend.abData[3] = fileId[1];
	stSend.abData[4] = 0x01; // file type
	stSend.abData[5] = AccessCond[0]; // access condition coding
	stSend.abData[6] = AccessCond[1];
	stSend.abData[7] = KeyNum[0]; // key number coding
	stSend.abData[8] = KeyNum[1];
	stSend.abData[9] = 0x00; // reserved for system
	stSend.abData[10] = 0x00;

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 16, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int WriteETFAndCheckErr(SCARDHANDLE hCard, BYTE *FileData, BYTE DataLen, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != WriteETFile(hCard, FileData, DataLen, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int WriteETFile(SCARDHANDLE hCardHandle, BYTE *FileData, BYTE DataLen, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xd6;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = DataLen;
	for(int iIndex=0; iIndex < DataLen; iIndex++)
	{
		stSend.abData[iIndex] = FileData[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5+DataLen, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int ReadETFAndCheckErr(SCARDHANDLE hCard, BYTE *FileData, BYTE DataLen, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != ReadETFile(hCard, FileData, DataLen, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int ReadETFile(SCARDHANDLE hCardHandle, BYTE *FileData, BYTE DataLen, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xb0;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = DataLen;

	// Transmit the request.
	lRecvBytes = 2+DataLen;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5, 
							NULL, stSend.abData, &lRecvBytes);

	if(SCARD_S_SUCCESS == (*pRet))
	{
		RecvStatus[0] = stSend.abData[lRecvBytes-2];
		RecvStatus[1] = stSend.abData[lRecvBytes-1];
		for(int iIndex=0; iIndex < DataLen; iIndex++)
		{
			FileData[iIndex] = stSend.abData[iIndex];
		}
	}
	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int CreateEFAndCheckErr(SCARDHANDLE hCard, BYTE *fileId, BYTE *FileSize, 
						BYTE *AccessCond, BYTE *KeyNum, BYTE RecLen, BYTE FileType,
						char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != CreateEFFile(hCard, fileId, FileSize, AccessCond, KeyNum, RecLen, FileType, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int CreateEFFile(SCARDHANDLE hCardHandle, BYTE *fileId, BYTE *FileSize, 
				 BYTE *AccessCond, BYTE *KeyNum, BYTE RecLen, 
				 BYTE FileType, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xe0;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x00;
	stSend.bP3 = 0x0B;
	stSend.abData[0] = FileSize[0];
	stSend.abData[1] = FileSize[1];
	stSend.abData[2] = fileId[0];
	stSend.abData[3] = fileId[1];
	stSend.abData[4] = FileType; // file type
	stSend.abData[5] = AccessCond[0]; // access condition coding
	stSend.abData[6] = AccessCond[1];
	stSend.abData[7] = KeyNum[0]; // key number coding
	stSend.abData[8] = KeyNum[1];
	stSend.abData[9] = RecLen;
	stSend.abData[10] = 0x00; // reserved for system or number of written recs

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 16, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int WriteEFRecAndCheckErr(SCARDHANDLE hCard, BYTE *FileData, BYTE DataLen, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != WriteEFRec(hCard, FileData, DataLen, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int WriteEFRec(SCARDHANDLE hCardHandle, BYTE *FileData, BYTE DataLen, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xd2;
	stSend.bP1 = 0x00;
	stSend.bP2 = 0x02;
	stSend.bP3 = DataLen;

	for(int iIndex=0; iIndex < DataLen; iIndex++)
	{
		stSend.abData[iIndex] = FileData[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5+DataLen, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int UpdateEFRecAndCheckErr(SCARDHANDLE hCard, BYTE *FileData, BYTE DataLen, BYTE RecNum, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != UpdateEFRec(hCard, FileData, DataLen, RecNum, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int UpdateEFRec(SCARDHANDLE hCardHandle, BYTE *FileData, BYTE DataLen, BYTE RecNum, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xdc;
	stSend.bP1 = RecNum;
	stSend.bP2 = 0x04; // absolute mode update
	stSend.bP3 = DataLen;

	for(int iIndex=0; iIndex < DataLen; iIndex++)
	{
		stSend.abData[iIndex] = FileData[iIndex];
	}

	// Transmit the request.
	lRecvBytes = 2;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5+DataLen, 
							NULL, (LPBYTE)RecvStatus, &lRecvBytes);

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

int ReadEFRecAndCheckErr(SCARDHANDLE hCard, BYTE *FileData, BYTE DataLen, BYTE RecNum, char *sMessage)
{
	BYTE RecvStatus[2];
	long lSysErr;

	if(0 != ReadEFRec(hCard, FileData, DataLen, RecNum, RecvStatus, &lSysErr))
	{
		CardErrToMessage(lSysErr, sMessage);
		return -1;
	}
	if (0 != InterpretFileOpStatusWord(RecvStatus, sMessage))
	{
		return -1;
	}
	return 0;
}

int ReadEFRec(SCARDHANDLE hCardHandle, BYTE *FileData, BYTE DataLen, BYTE RecNum, BYTE *RecvStatus, long *pRet)
{
	APDUStruct stSend;
	unsigned long lRecvBytes;

	stSend.bCla = 0x00;
	stSend.bIns = 0xb2;
	stSend.bP1 = RecNum;
	stSend.bP2 = 0x04; // absolute mode read
	stSend.bP3 = DataLen;

	// Transmit the request.
	lRecvBytes = 2+DataLen;
	*pRet = SCardTransmit(hCardHandle, SCARD_PCI_T0, (LPCBYTE)&stSend, 5, 
							NULL, (LPBYTE)stSend.abData, &lRecvBytes);


	if(SCARD_S_SUCCESS == (*pRet))
	{
		RecvStatus[0] = stSend.abData[lRecvBytes-2];
		RecvStatus[1] = stSend.abData[lRecvBytes-1];
		for(int iIndex=0; iIndex < DataLen; iIndex++)
		{
			FileData[iIndex] = stSend.abData[iIndex];
		}
	}

	return (SCARD_S_SUCCESS == (*pRet)) ? 0 : -1;
}

