/*
 * @(#)SCMessageTool.cpp	1.00 10th June 2000
 * 
 * Functions that take care of translating error codes for both PC/SC errors
 * and smartcard errors into message strings.
 *
 * Modification Log:
 */

/**
 *
 * SCTool is a sample application using Schlumberger PayFlex 1K smart cards.
 * Using this application you can personalize a card, and store data either
 * protected with a pin number or without protection, as you desire.
 * 
 * Several generic functions have been written for SmartCards, so that they
 * can be ported later to any type of SmartCard. 
 * 
 * For more information on SmartCards and how to program them, refer to the
 * "Smart Card Crash Course" at my website: 
 * http://tanmaykm.tripod.com.
 *
 * You are free to use this code and to make modifications provided
 * this notice is retained.
 * <p>
 * If you found this useful, please add a note of acknowledgement to my guestbook,
 * or send me a mail at <a href="mailto:tanmaykm@hotmail.com">tanmaykm@hotmail.com</a>. 
 * If you would like to report a bug or suggest some improvements, 
 * you are most welcome. I will be happy to help you use this piece of code.
 * <p>
 *
 * @author 	    Tanmay K. Mohapatra
 * @version     1.00, 10th June, 2000
 */

#include <windows.h>
#include <winscard.h>
#include <scarderr.h>
#include <stdio.h>

#include "SCMessageTool.hxx"


void PrintCardStatus(SCARDHANDLE hCardHandle)
{
	char            szReader[256], sMsg[256];
	DWORD           cch = 256;
	BYTE            bAttr[32];
	DWORD           cByte = 32, cByteIndex;
	DWORD           dwState, dwProtocol;
	LONG            lReturn;

	// Determine the status.
	// hCardHandle was set by an earlier call to SCardConnect.
	lReturn = SCardStatus(hCardHandle, szReader, &cch, &dwState, &dwProtocol,
						  (LPBYTE)&bAttr, &cByte); 

	if(SCARD_S_SUCCESS != lReturn)
	{
		CardErrToMessage(lReturn, sMsg);
		printf("Failed SCardStatus\n%s\n", sMsg);
		return;
	}

	// Examine retrieved status elements.
	// Here, we'll look at the reader name and card state.
	printf("Reader: %s\n", szReader);
	printf("Card Status: ");
	switch(dwState)
	{
		case SCARD_ABSENT:
			printf("Card absent.\n");
			break;
		case SCARD_PRESENT:
			printf("There is a card in the reader, but it has not been moved into position for use.\n");
			break;
		case SCARD_SWALLOWED:
			printf("There is a card in the reader in position for use. The card is not powered.\n");
			break;
		case SCARD_POWERED:
			printf("Power is being provided to the card, but the reader driver is unaware of the mode of the card.\n");
			break;
		case SCARD_NEGOTIABLE:
			printf("Card reset and waiting PTS negotiation.\n");
			break;
		case SCARD_SPECIFIC:
			printf("The card has been reset and specific communication protocols have been established.\n");
			break;
		default:
			printf("Unknown or unexpected card state.\n");
			break;
	}
	printf("Protocol: ");
	switch (dwProtocol)
	{
		case SCARD_PROTOCOL_RAW:
			printf("Raw\n");
			break;
		case SCARD_PROTOCOL_T0:
			printf("T0\n");
			break;
		case SCARD_PROTOCOL_T1:
			printf("T1\n");
			break;
		default:
			printf("Unknown or unexpected protocol.\n");
			break;
	}
	printf("ATR: %d bytes, ", cByte);
	for(cByteIndex=0; cByteIndex < cByte; cByteIndex++)
	{
		printf("%2.2xh ", bAttr[cByteIndex]);
	}
	printf("\n");

	printf("Interpreting ATR:\n");
	printf("Payflex Id:\n");
	printf("\t%2.2xh : RFU\n", bAttr[9]);
	printf("\t%2.2xh : Certification Mode\n", bAttr[10]);
	printf("\t%2.2xh : EEPROM Size\n", bAttr[11]);
	printf("\t%2.2xh : Card Identification\n", bAttr[12]);

	// print card type
	if( (0 == bAttr[9]) && (0x01 == bAttr[10]) )
	{
		printf("Payflex");
		printf(" %xK", bAttr[11]);
		switch(bAttr[12])
		{
			case 0xA9: printf(" User"); break;
			case 0xc9: printf(" SAM"); break;
		}
	}
	else
	{
		printf("Unknown");
	}
	printf(" Card\n");
}


int InterpretFileOpStatusWord(BYTE *RecvStatus, char *sMsg)
{
	int iSuccess = -1;
	if((0x6e == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "Class != 0");
	}
	else if((0x6d == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "Unknown instruction");
	}
	else if((0x6a == RecvStatus[0]) && (0x80 == RecvStatus[1]))
	{
		sprintf(sMsg, "File ID to create already existing in the parent DF; OR"
					  " BS/RL is not an integer value or exceeds 256; OR"
					  " Other improper size combinations.");
	}
	else if((0x6a == RecvStatus[0]) && (0x82 == RecvStatus[1]))
	{
		sprintf(sMsg, "File ID not found or Father DF invalid => can't select EF; or Checksum is incorrect");
	}
	else if((0x6a == RecvStatus[0]) && (0x83 == RecvStatus[1]))
	{
		sprintf(sMsg, "All records already written");
	}
	else if((0x6a == RecvStatus[0]) && (0x84 == RecvStatus[1]))
	{
		sprintf(sMsg, "Insufficient memory space available");
	}
	else if((0x6a == RecvStatus[0]) && (0x86 == RecvStatus[1]))
	{
		sprintf(sMsg, "Incorrect P1 or P2");
	}
	else if((0x6b == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "Offset out of the EF");
	}
	else if((0x62 == RecvStatus[0]) && (0x83 == RecvStatus[1]))
	{
		sprintf(sMsg, "Selected DF File is invalid");
	}
	else if((0x62 == RecvStatus[0]) && (0x82 == RecvStatus[1]))
	{
		sprintf(sMsg, "File ID not found");
	}
	else if((0x62 == RecvStatus[0]) && (0x81 == RecvStatus[1]))
	{
		sprintf(sMsg, "Selected EPF is corrupted");
	}
	else if((0x63 == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "Wrong data, KEY is not blocked yet");
	}
	else if((0x67 == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "P3 != 02");
	}
	else if((0x69 == RecvStatus[0]) && (0x82 == RecvStatus[1]))
	{
		sprintf(sMsg, "AC are not fulfilled");
	}
	else if((0x69 == RecvStatus[0]) && (0x83 == RecvStatus[1]))
	{
		sprintf(sMsg, "Wrong CHV/KEY. KEY is blocked");
	}
	else if((0x69 == RecvStatus[0]) && (0x85 == RecvStatus[1]))
	{
		sprintf(sMsg, "File type selected is improper.");
	}
	else if((0x90 == RecvStatus[0]) && (0x00 == RecvStatus[1]))
	{
		sprintf(sMsg, "File is selected");
		iSuccess = 0;
	}
	else
	{
		sprintf(sMsg, "Unknown error[%2.2xh %2.2xh]", RecvStatus[0], RecvStatus[1]);
	}

	return iSuccess;
}


void CardErrToMessage(HRESULT hr, char *sMsg)
{
	switch(hr)
	{
#ifdef SCARD_E_BAD_SEEK
	case SCARD_E_BAD_SEEK:
		sprintf(sMsg, "An error occurred in setting the smart card file object pointer.\n");
		break;
#endif
	case SCARD_E_CANCELLED:
		sprintf(sMsg, "The action was canceled by an SCardCancel request.");
		break;
	case SCARD_E_CANT_DISPOSE:
		sprintf(sMsg, "The system could not dispose of the media in the requested manner.");
		break;
	case SCARD_E_CARD_UNSUPPORTED:
		sprintf(sMsg, "The smart card does not meet minimal requirements for support.");
		break;
#ifdef SCARD_E_CERTIFICATE_UNAVAILABLE
	case SCARD_E_CERTIFICATE_UNAVAILABLE:
		sprintf(sMsg, "The requested certificate could not be obtained.");
		break;
#endif
#ifdef SCARD_E_COMM_DATA_LOST
	case SCARD_E_COMM_DATA_LOST:
		sprintf(sMsg, "A communications error with the smart card has been detected.");
		break;
#endif
#ifdef SCARD_E_DIR_NOT_FOUND
	case SCARD_E_DIR_NOT_FOUND:
		sprintf(sMsg, "The specified directory does not exist in the smart card.");
		break;
#endif
	case SCARD_E_DUPLICATE_READER:
		sprintf(sMsg, "The reader driver didn't produce a unique reader name.");
		break;
#ifdef SCARD_E_FILE_NOT_FOUND
	case SCARD_E_FILE_NOT_FOUND:
		sprintf(sMsg, "The specified file does not exist in the smart card.");
		break;
#endif
#ifdef SCARD_E_ICC_CREATEORDER
	case SCARD_E_ICC_CREATEORDER:
		sprintf(sMsg, "The requested order of object creation is not supported.");
		break;
#endif
#ifdef SCARD_E_ICC_INSTALLATION
	case SCARD_E_ICC_INSTALLATION:
		sprintf(sMsg, "No primary provider can be found for the smart card.");
		break;
#endif
	case SCARD_E_INSUFFICIENT_BUFFER:
		sprintf(sMsg, "The data buffer for returned data is too small for the returned data.");
		break;
	case SCARD_E_INVALID_ATR:
		sprintf(sMsg, "An ATR string obtained from the registry is not a valid ATR string.");
		break;
#ifdef SCARD_E_INVALID_CHV
	case SCARD_E_INVALID_CHV:
		sprintf(sMsg, "The supplied PIN is incorrect.");
		break;
#endif
	case SCARD_E_INVALID_HANDLE:
		sprintf(sMsg, "The supplied handle was invalid.");
		break;
	case SCARD_E_INVALID_PARAMETER:
		sprintf(sMsg, "One or more of the supplied parameters could not be properly interpreted.");
		break;
	case SCARD_E_INVALID_TARGET:
		sprintf(sMsg, "Registry startup information is missing or invalid.");
		break;
	case SCARD_E_INVALID_VALUE:
		sprintf(sMsg, "One or more of the supplied parameter values could not be properly interpreted.");
		break;
#ifdef SCARD_E_NO_ACCESS
	case SCARD_E_NO_ACCESS:
		sprintf(sMsg, "Access is denied to the file.");
		break;
#endif
#ifdef SCARD_E_NO_DIR
	case SCARD_E_NO_DIR:
		sprintf(sMsg, "The supplied path does not represent a smart card directory.");
		break;
#endif
#ifdef SCARD_E_NO_FILE
	case SCARD_E_NO_FILE:
		sprintf(sMsg, "The supplied path does not represent a smart card file.");
		break;
#endif
	case SCARD_E_NO_MEMORY:
		sprintf(sMsg, "Not enough memory available to complete this command.");
		break;
#ifdef SCARD_E_NO_READERS_AVAILABLE
	case SCARD_E_NO_READERS_AVAILABLE:
		sprintf(sMsg, "No smart card reader is available.");
		break;
#endif
	case SCARD_E_NO_SERVICE:
		sprintf(sMsg, "The smart card resource manager is not running.");
		break;
	case SCARD_E_NO_SMARTCARD:
		sprintf(sMsg, "The operation requires a smart card, but no smart card is currently in the device.");
		break;
#ifdef SCARD_E_NO_SUCH_CERTIFICATE
	case SCARD_E_NO_SUCH_CERTIFICATE:
		sprintf(sMsg, "The requested certificate does not exist.");
		break;
#endif
	case SCARD_E_NOT_READY:
		sprintf(sMsg, "The reader or card is not ready to accept commands.");
		break;
	case SCARD_E_NOT_TRANSACTED:
		sprintf(sMsg, "An attempt was made to end a non-existent transaction.");
		break;
	case SCARD_E_PCI_TOO_SMALL:
		sprintf(sMsg, "The PCI receive buffer was too small.");
		break;
	case SCARD_E_PROTO_MISMATCH:
		sprintf(sMsg, "The requested protocols are incompatible with the protocol currently in use with the card.");
		break;
	case SCARD_E_READER_UNAVAILABLE:
		sprintf(sMsg, "The specified reader is not currently available for use.");
		break;
	case SCARD_E_READER_UNSUPPORTED:
		sprintf(sMsg, "The reader driver does not meet minimal requirements for support.");
		break;
	case SCARD_E_SERVICE_STOPPED:
		sprintf(sMsg, "The smart card resource manager has shut down.");
		break;
	case SCARD_E_SHARING_VIOLATION:
		sprintf(sMsg, "The smart card cannot be accessed because of other outstanding connections.");
		break;
	case SCARD_E_SYSTEM_CANCELLED:
		sprintf(sMsg, "The action was canceled by the system, presumably to log off or shut down.");
		break;
	case SCARD_E_TIMEOUT:
		sprintf(sMsg, "The user-specified timeout value has expired.");
		break;
#ifdef SCARD_E_UNEXPECTED
	case SCARD_E_UNEXPECTED:
		sprintf(sMsg, "An unexpected card error has occurred.");
		break;
#endif
	case SCARD_E_UNKNOWN_CARD:
		sprintf(sMsg, "The specified smart card name is not recognized.");
		break;
	case SCARD_E_UNKNOWN_READER:
		sprintf(sMsg, "The specified reader name is not recognized.");
		break;
#ifdef SCARD_E_UNKNOWN_RES_MNG
	case SCARD_E_UNKNOWN_RES_MNG:
		sprintf(sMsg, "An unrecognized error code was returned from a layered component.");
		break;
#endif
#ifdef SCARD_E_UNSUPPORTED_FEATURE
	case SCARD_E_UNSUPPORTED_FEATURE:
		sprintf(sMsg, "This smart card does not support the requested feature.");
		break;
#endif
#ifdef SCARD_E_WRITE_TOO_MANY
	case SCARD_E_WRITE_TOO_MANY:
		sprintf(sMsg, "An attempt was made to write more data than would fit in the target object.");
		break;
#endif
	case SCARD_F_COMM_ERROR:
		sprintf(sMsg, "An internal communications error has been detected.");
		break;
	case SCARD_F_INTERNAL_ERROR:
		sprintf(sMsg, "An internal consistency check failed.");
		break;
	case SCARD_F_UNKNOWN_ERROR:
		sprintf(sMsg, "An internal error has been detected, but the source is unknown.");
		break;
	case SCARD_F_WAITED_TOO_LONG:
		sprintf(sMsg, "An internal consistency timer has expired.");
		break;
	case SCARD_P_SHUTDOWN:
		sprintf(sMsg, "The operation has been aborted to allow the server application to exit.");
		break;
	case SCARD_S_SUCCESS:
		sprintf(sMsg, "No error was encountered.");
		break;
#ifdef SCARD_W_CANCELLED_BY_USER
	case SCARD_W_CANCELLED_BY_USER:
		sprintf(sMsg, "The action was cancelled by the user.");
		break;
#endif
#ifdef SCARD_W_CHV_BLOCKED
	case SCARD_W_CHV_BLOCKED:
		sprintf(sMsg, "The card cannot be accessed because the maximum number of PIN entry attempts has been reached.");
		break;
#endif
#ifdef SCARD_W_EOF
	case SCARD_W_EOF:
		sprintf(sMsg, "The end of the smart card file has been reached.");
		break;
#endif
	case SCARD_W_REMOVED_CARD:
		sprintf(sMsg, "The smart card has been removed, so that further communication is not possible.");
		break;
	case SCARD_W_RESET_CARD:
		sprintf(sMsg, "The smart card has been reset, so any shared state information is invalid.");
		break;
#ifdef SCARD_W_SECURITY_VIOLATION
	case SCARD_W_SECURITY_VIOLATION:
		sprintf(sMsg, "Access was denied because of a security violation.");
		break;
#endif
	case SCARD_W_UNPOWERED_CARD:
		sprintf(sMsg, "Power has been removed from the smart card, so that further communication is not possible.");
		break;
	case SCARD_W_UNRESPONSIVE_CARD:
		sprintf(sMsg, "The smart card is not responding to a reset.");
		break;
	case SCARD_W_UNSUPPORTED_CARD:
		sprintf(sMsg, "The reader cannot communicate with the card, due to ATR string configuration conflicts.");
		break;
#ifdef SCARD_W_WRONG_CHV
	case SCARD_W_WRONG_CHV:
		sprintf(sMsg, "The card cannot be accessed because the wrong PIN was presented.");
		break;
#endif
	default:
		{
			LPVOID lpMsgBuf;
			FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&lpMsgBuf, 0, NULL);
			sprintf(sMsg, "[%ld] %s", hr, lpMsgBuf);
			LocalFree(lpMsgBuf);
		}
		break;
	}
}

